from unittest2 import TestCase
from mock import patch, Mock, sentinel

from exam.helpers import intercept, rm_f, track, mock_import
from exam.decorators import fixture

from describe import expect


@patch('exam.helpers.shutil')
class TestRmrf(TestCase):

    path = '/path/to/folder'

    def test_calls_shutil_rmtreee(self, shutil):
        rm_f(self.path)
        shutil.rmtree.assert_called_once_with(self.path, ignore_errors=True)

    @patch('exam.helpers.os')
    def test_on_os_errors_calls_os_remove(self, os, shutil):
        shutil.rmtree.side_effect = OSError
        rm_f(self.path)
        os.remove.assert_called_once_with(self.path)


class TestTrack(TestCase):

    @fixture
    def foo_mock(self):
        return Mock()

    @fixture
    def bar_mock(self):
        return Mock()

    def test_makes_new_mock_and_attaches_each_kwarg_to_it(self):
        tracker = track(foo=self.foo_mock, bar=self.bar_mock)
        expect(tracker.foo).to == self.foo_mock
        expect(tracker.bar).to == self.bar_mock


class TestMockImport(TestCase):

    def test_is_a_context_manager_that_yields_patched_import(self):
        with mock_import('foo') as mock_foo:
            import foo
            expect(foo).to == mock_foo

    def test_mocks_import_for_packages(self):
        with mock_import('foo.bar.baz') as mock_baz:
            import foo.bar.baz
            expect(foo.bar.baz).to == mock_baz

    @mock_import('foo')
    def test_can_be_used_as_a_decorator_too(self, mock_foo):
        import foo
        expect(foo).to == mock_foo

    @mock_import('foo')
    @mock_import('bar')
    def test_stacked_adds_args_bottom_up(self, mock_bar, mock_foo):
        import foo
        import bar
        expect(mock_bar).to == bar
        expect(mock_foo).to == foo


class TestIntercept(TestCase):

    class Example(object):
        def method(self, positional, keyword):
            return sentinel.METHOD_RESULT

    def test_intercept(self):
        ex = self.Example()

        def counter(positional, keyword):
            assert positional is sentinel.POSITIONAL_ARGUMENT
            assert keyword is sentinel.KEYWORD_ARGUMENT
            result = yield
            assert result is sentinel.METHOD_RESULT
            counter.calls += 1

        counter.calls = 0

        intercept(ex, 'method', counter)
        expect(counter.calls).to == 0
        assert ex.method(sentinel.POSITIONAL_ARGUMENT,
            keyword=sentinel.KEYWORD_ARGUMENT) is sentinel.METHOD_RESULT
        expect(counter.calls).to == 1

        ex.method.unwrap()
        assert ex.method(sentinel.POSITIONAL_ARGUMENT,
            keyword=sentinel.KEYWORD_ARGUMENT) is sentinel.METHOD_RESULT
        expect(counter.calls).to == 1



from unittest2 import TestCase

from exam import Exam, fixture
from exam.asserts import AssertsMixin

from describe import expect


class AssertChangesMixin(Exam, TestCase):

    case = fixture(AssertsMixin)
    thing = fixture(list)

    def no_op_context(self, *args, **kwargs):
        with self.case.assertChanges(len, self.thing, *args, **kwargs):
            pass

    def test_checks_change_on_callable_passed(self):
        with self.case.assertChanges(len, self.thing, before=0, after=1):
            self.thing.append(1)

    def test_after_check_asserts_ends_on_after_value(self):
        self.thing.append(1)
        with self.case.assertChanges(len, self.thing, after=2):
            self.thing.append(1)

    def test_before_check_asserts_starts_on_before_value(self):
        self.thing.append(1)
        with self.case.assertChanges(len, self.thing, before=1):
            self.thing.append(1)
            self.thing.append(2)

    def test_verifies_value_must_change_no_matter_what(self):
        self.thing.append(1)

        with expect.to_raise_error(AssertionError):
            self.no_op_context(after=1)

        with expect.to_raise_error(AssertionError):
            self.no_op_context(before=1)

        with expect.to_raise_error(AssertionError):
            self.no_op_context()



from mock import sentinel
from unittest2 import TestCase
import itertools

from exam.decorators import fixture, before, after, around, patcher
from exam.cases import Exam

from describe import expect

from dummy import get_thing, get_it


class SimpleTestCase(object):

    def __init__(self):
        self.cleanups = []
        self.setups = 0
        self.teardowns = 0

    def setUp(self):
        self.setups += 1

    def tearDown(self):
        self.teardowns += 1

    def run(self, *args, **kwargs):
        self.state_when_run = list(self.calls)

    def addCleanup(self, func):
        self.cleanups.append(func)


class MyTestCase(Exam, SimpleTestCase):

    @patcher('tests.dummy.thing')
    def dummy_thing(self):
        return sentinel.mock

    dummy_it = patcher('tests.dummy.it', return_value=12)

    def __init__(self):
        self.calls = []
        super(MyTestCase, self).__init__()

    @before
    def append_wine(self):
        self.calls.append('merlot')

    @before
    def append_self(self):
        self.calls.append(type(self))

    @after
    def append_number(self):
        self.calls.append('2 in parent class')

    @after
    def append_person(self):
        self.calls.append('jeff')

    @around
    def fairy_tail(self):
        self.calls.append('hansel')
        yield
        self.calls.append('gretle')


class ExtendedTestCase(MyTestCase):

    @before
    def append_cheese(self):
        self.calls.append('chedder')

    @after
    def append_meat(self):
        self.calls.append('salami')

    @around
    def movies(self):
        self.calls.append('turner')
        yield
        self.calls.append('hootch')

    @before
    def append_self(self):
        self.calls.append(type(self))

    @after
    def append_number(self):
        self.calls.append('two in subclass')


# TODO: Make the subclass checking just be a subclass of the test case
class TestExam(Exam, TestCase):

    @fixture
    def case(self):
        return MyTestCase()

    @fixture
    def subclassed_case(self):
        return ExtendedTestCase()

    @after
    def stop_patchers(self):
        cleanups = (self.case.cleanups, self.subclassed_case.cleanups)

        for cleanup in itertools.chain(*cleanups):
            if hasattr(cleanup.im_self, 'is_local'):  # Is the mock started?
                cleanup()

    @property
    def other_thing(self):
        return get_thing()

    @property
    def other_it(self):
        return get_it()

    def test_assert_changes_is_asserts_mixin_assert_changes(self):
        from exam.asserts import AssertsMixin
        expect(AssertsMixin.assertChanges, Exam.assertChanges)

    def test_before_adds_each_method_to_set_up(self):
        expect(self.case.calls).to == []
        self.case.setUp()
        expect(self.case.calls).to == [MyTestCase, 'merlot']

    def test_after_adds_each_method_to_tear_down(self):
        expect(self.case.calls).to == []
        self.case.tearDown()
        expect(self.case.calls).to == ['jeff', '2 in parent class']

    def test_around_calls_methods_before_and_after_run(self):
        expect(self.case.calls).to == []
        self.case.run()
        expect(self.case.state_when_run).to == ['hansel']
        expect(self.case.calls).to == ['hansel', 'gretle']

    def test_before_works_on_subclasses(self):
        expect(self.subclassed_case.calls).to == []
        self.subclassed_case.setUp()

        # The only concern with ordering here is that the parent class's @before
        # hook fired before it's parents.  The actual order of the @before hooks
        # within a level of class is irrelevant.
        expect(self.subclassed_case.calls[0]).to == 'merlot'
        expect(sorted(self.subclassed_case.calls[1:])).to == ['chedder', ExtendedTestCase]

    def test_after_works_on_subclasses(self):
        expect(self.subclassed_case.calls).to == []
        self.subclassed_case.tearDown()
        expect(self.subclassed_case.calls).to == ['jeff', 'salami', 'two in subclass']

    def test_around_works_with_subclasses(self):
        expect(self.subclassed_case.calls).to == []
        self.subclassed_case.run()
        expect(self.subclassed_case.state_when_run).to == ['hansel', 'turner']
        expect(self.subclassed_case.calls).to == ['hansel', 'turner', 'hootch', 'gretle']

    def test_patcher_start_value_is_added_to_case_dict_on_setup(self):
        self.case.setUp()
        expect(self.case.dummy_thing).to == sentinel.mock

    def test_patcher_patches_object_on_setup_and_adds_patcher_to_cleanup(self):
        expect(self.other_thing).to != sentinel.mock
        self.case.setUp()
        expect(self.other_thing).to == sentinel.mock
        [cleanup() for cleanup in self.case.cleanups]
        expect(self.other_thing).to != sentinel.mock

    def test_patcher_lifecycle_works_on_subclasses(self):
        expect(self.other_thing).to != sentinel.mock
        self.subclassed_case.setUp()
        expect(self.other_thing).to == sentinel.mock
        [cleanup() for cleanup in self.subclassed_case.cleanups]
        expect(self.other_thing).to != sentinel.mock

    def test_patcher_patches_with_a_magic_mock_if_no_function_decorated(self):
        expect(self.other_it()).to != 12
        self.case.setUp()
        expect(self.other_it()).to == 12
        self.case.cleanups[0]()
        expect(self.other_thing).to != 12

    def test_calls_super_setup(self):
        expect(self.case.setups).to == 0
        self.case.setUp()
        expect(self.case.setups).to == 1

    def test_calls_super_teardown(self):
        expect(self.case.teardowns).to == 0
        self.case.tearDown()
        expect(self.case.teardowns).to == 1



IRRELEVANT = object()


class ChangeWatcher(object):

    def __init__(self, thing, *args, **kwargs):
        self.thing = thing
        self.args = args
        self.kwargs = kwargs
        self.expected_before = kwargs.pop('before', IRRELEVANT)
        self.expected_after = kwargs.pop('after', IRRELEVANT)

    def __enter__(self):
        self.before = self.__apply()

        if not self.expected_before is IRRELEVANT:
            check = self.before == self.expected_before
            assert check, self.__precondition_failure_msg_for('before')

    def __exit__(self, type, value, traceback):
        self.after = self.__apply()

        if not self.expected_after is IRRELEVANT:
            check = self.after == self.expected_after
            assert check, self.__precondition_failure_msg_for('after')

        assert self.before != self.after, self.__equality_failure_message

    def __apply(self):
        return self.thing(*self.args, **self.kwargs)

    @property
    def __equality_failure_message(self):
        return 'Expected before %r != %r after' % (self.before, self.after)

    def __precondition_failure_msg_for(self, condition):
        return '%s value did not change (%s)' % (
            condition,
            getattr(self, condition)
        )


class AssertsMixin(object):
    assertChanges = ChangeWatcher
